/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDLOctetStream.h"
#import "IDLValueType_Private.h"
#import "IDLDefinitionScope.h"
#import "IDLObject.h"

@implementation IDLOctetStream

NSString* IDLValueEndException = @"IDLValueEndException";

static BOOL _encodesWithChunks = YES;

- (void) dealloc
{
    [_ref release];
    [super dealloc];
}

- (long) _decodeLong
{
    // No alignment
    NSRange r = {_cursor, sizeof(long)};
    long value;
    [_data getBytes:&value range:r];
    _cursor += r.length;
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)&value length:sizeof(long)];
    return value;
}

- (void) _encodeLong:(long)val
{
    long value = val;
    [_data appendBytes:&value length:sizeof(long)];
}

- (long) decodeLongAt:(long)location
{
    NSRange r = {location, sizeof(long)};
    long value;
    [_data getBytes:&value range:r];
    if(_flags._shouldSwap)
        [self _swapBytes:(unsigned char*)&value length:sizeof(long)];
    return value;
}

- (void) encodeLong:(long)val at:(long)location
{
    NSRange r = {location, sizeof(long)};
    long value = val;
    [_data replaceBytesInRange:r withBytes:&value];
}
    
- (long) _peepTag
{
    long location = _cursor;
    long tag;
    tag = [self _decodeLong];
    _cursor = location;
    return tag;
}

- (void) decodeChunk
{
    long tag;
    [super _alignCursorTo:sizeof(long)];
    tag = [self _peepTag];
    
    if(tag > 0 && tag < 0x7fffff00){
        _cursor += sizeof(long);
        _chunkBreak = _cursor + tag;
        _flags.inChunk = YES;
    }else if(tag >= 0x7fffff00){
		_flags.inChunk = NO;
    }else if(tag < 0)
        [NSException raise:IDLValueEndException format:@"value-end-tag"];
    else{// tag == 0
        [NSException raise:CORBAMarshalException format:@"invalid value tag (0)"];
    }
}

- (void) closeChunk
{
    if(_flags.inChunk){
        _flags.inChunk = NO;
        if(_flags._isOutStream){
            long l = [_data length] - _chunkBreak - sizeof(long);
            [self encodeLong:l at:_chunkBreak];
        }else // Input stream
            _cursor = _chunkBreak;
    }
}

- (void) beginChunk
{
    if(_flags.useChunks){
        if(!_flags.inChunk){
            [super _alignTo:sizeof(long)];
            _chunkBreak = [_data length];
            [self _encodeLong:0];
            _flags.inChunk = YES;
        }
    }
}

- (void) _setUseChunks:(BOOL)flag
{
    if(_valueLevel == 0)
        _flags.useChunks = flag?YES:NO;
}

- (BOOL) _useChunks
{
    return (_valueLevel > 0) && _flags.useChunks;
}

- (void) _alignTo:(unsigned int)b
{
    if(_flags._needsAlign){ // This cannot be called when in chunk
        int length = [_data length] + _flags._offset;
        int l = length + 7 - (length+7)%8;
        if(l > length)
            [_data increaseLengthBy:(l-length)];
        _flags._needsAlign = NO;
    }

    if([self _useChunks] && !_flags.inChunk)
        [self beginChunk];
    
    if(b > 1){
        int length = [_data length] + _flags._offset;
        int l = length + b - 1 - (length+b-1)%b;
        if(l > length)
            [_data increaseLengthBy:(l-length)];
    }
}


- (void) _alignCursorTo:(unsigned int)b
{
    if(_flags._needsAlign){ // This cannot be inside chunk
        long l = _cursor - (_cursor + _flags._offset)%(b=8);
        if(l < _cursor)
            l += b;
        _cursor = l;
        _flags._needsAlign = NO;
        if(_cursor >= [_data length])
            [NSException raise:CORBAMarshalException format:@"cusror is beyond the boundary of stream"];
        return;
    }
    
    if([self _useChunks] && !_flags.inChunk)
        [self decodeChunk];
    
    if(b > 1){
        long l = _cursor - (_cursor + _flags._offset)%b;
        if(l < _cursor)
            l += b;
        
        if(_flags.inChunk && l >= _chunkBreak){
            _flags.inChunk = NO;
            [self decodeChunk];
            l = _cursor - (_cursor + _flags._offset)%b;
            if(l < _cursor)
                l += b;
        }
        
        _cursor = l;
        if(_cursor >= [_data length])
            [NSException raise:CORBAMarshalException format:@"cusror is beyond the boundary of stream"];
    }
}

- (void) registerObject:object withLocation:(long)location
{
    NSString* key = [NSString stringWithFormat:@"l=0x%x",location];
    if(!_ref)
        _ref = [[NSMutableDictionary allocWithZone:[self zone]]init];
    [_ref setObject:object forKey:key];
}

- (void) registerLocation:(long)location forObject:object
{
    if(!_ref)
        _ref = [[NSMutableDictionary allocWithZone:[self zone]]init];
    if(object){
        NSString* key = [NSString stringWithFormat:@"0x%x", object];
        [_ref setObject:[NSNumber numberWithLong:location] forKey:key];
    }
}

- (long) locationForObject:object
{
    if(object){
        NSString* key = [NSString stringWithFormat:@"0x%x", object];
        return [[_ref objectForKey:key]longValue];
    }else
        return 0;
}

- objectWithLocation:(long)location
{
    NSString* key = [NSString stringWithFormat:@"l=0x%x",location];
    return [_ref objectForKey:key];
}

- (NSString*) decodeStringWithIndirection
{
    NSRange r = {_cursor,sizeof(long)};
    long location;
    long length;
    [super _alignCursorTo:sizeof(long)];
    location = _cursor;
    length = [self _decodeLong];
    r.location = _cursor;
    
    if(r.location + length > [_data length])
        [NSException raise:CORBAMarshalException 
                format:@"string length beyond buffer"];

    if(length == -1){ //indirection
        long offset = [self _decodeLong];
        return [self objectWithLocation:r.location + offset];
    }else
    if(length > 1){ // length = 0 ==> <nil>
		r.length = length - 1; // Do not include terminating 0
		NSData* bytes = [_data subdataWithRange:r];
        _cursor = r.location + length;
		return [[[NSString allocWithZone:[self zone]]initWithData:bytes encoding:[self stringEncoding]]autorelease];
	}else
	if(length == 1){
		_cursor = r.location + length;
		return VXDecodeEmptyStringAsNil? nil : @"";
	}
	else
    if(length < 0)
        [NSException raise:@"CORBAMarshalExeption" 
                format:@"negative string length"];
    return nil;
}

- (void) _encodeString:(NSString*)string
{
	NSData* bytes = [string dataUsingEncoding:[self stringEncoding] allowLossyConversion:NO];
    long len = [bytes length] + 1;
	char zero = 0;
    [super _alignTo:sizeof(long)];
    [_data appendBytes:&len length:sizeof(long)];
    [_data appendData:bytes];
	[_data appendBytes:&zero length:1];
}

- (NSArray*) decodeStringSequence
{
    NSMutableArray* array;
    long count;
    long location;
    
    [super _alignCursorTo:sizeof(long)];
    location = _cursor;
    count = [self _decodeLong];
    
    if(count == -1){
        long offset = [self _decodeLong];
        if(offset < 0 && location + sizeof(long) + offset > 0)
            return [self objectWithLocation:location + sizeof(long) + offset];
        else
            [NSException raise:CORBAMarshalException format:@"invalid indirection"];
    }
    array = [NSMutableArray array];
    while(count-- > 0){
        NSString* item = [self decodeStringWithIndirection];
        if(item)
            [array addObject:item];
    }
    [self registerObject:array withLocation:location];
    return array;
}

- (void) decodeValue:(id*)obj forObjectCoder:(IDLValueType*)valuetype
{
    long tag = 0;
    long location;
    
    [super _alignCursorTo:sizeof(long)];
    if(_flags.inChunk && _cursor >= _chunkBreak)
        _flags.inChunk = NO;
    location = _cursor;
    tag = [self _decodeLong];
    
    if(!_flags.inChunk && tag > 0 && tag < 0x7fffff00){ // Chunk tag
        _cursor -= sizeof(long);
        [self decodeChunk];
        location = _cursor;
        tag = [self _decodeLong];
    }
    
	// value-end tag can be present if and only if useChunks == YES
    if((![self _useChunks] || _flags.inChunk) && tag == -1){ // indirection
        long offset = [self _decodeLong];
        if(offset < 0 && location + sizeof(long) + offset > 0)
            *obj = [self objectWithLocation:location + sizeof(long) + offset];
        else
            [NSException raise:CORBAMarshalException format:@"invalid indirection"];
        return;
    }else
    if((![self _useChunks] || _flags.inChunk) && tag == 0){ // nil value
        *obj = nil;
    }else
    if([self _useChunks] && !_flags.inChunk && tag < 0){ // value-end tag
        // This should not happen ???
        // Try to handle it anyway
        
        // The outer value has level '_valueLevel'
        if(tag + _valueLevel < 0){ // pevious nested value did not take it ?
            [NSException raise:CORBAMarshalException 
                format:@"expected value tag, read value-end tag '%d' with outer value level = %d",
                    tag, _valueLevel];
        }else // outer value ended
            [NSException raise:IDLValueEndException format:@"unexpected value-end-tag"];
        
    }else
    if(!_flags.inChunk && tag >= 0x7fffff00){ //chunk, if any, should have ended.
        NSString* codebase = nil;
        IDLValueType* valueDef = valuetype;
        NSString* className;
        NSString* typeId = nil;
        id class = nil;
        
        if(tag&0x01)
            codebase = [self decodeStringWithIndirection];
        
        if(tag&0x06 == 0x06){
            NSArray* array = [self decodeStringSequence];
            typeId = [array objectAtIndex:0];
            valueDef = (IDLValueType*)[[IDLDefinitionScope currentScope]definitionForTypeId:typeId];
        }else
        if(tag&0x02){
            typeId = [self decodeStringWithIndirection];
            valueDef = (IDLValueType*)[[IDLDefinitionScope currentScope]definitionForTypeId:typeId];
        }
        
        if(!valueDef)
            [NSException raise:CORBAMarshalException 
                    format:@"cannot locate definition for typeID '%@'", typeId];
        className = [valueDef objectClassName];
        if(className)
            class = NSClassFromString(className);
        if(!class)
            class = [IDLValueType defaultObjectClass];
        if(!class)
            class = [IDLObject class];
        
		if([class instancesRespondToSelector:@selector(initWithIDLType:)])
			*obj = [[[class alloc]initWithIDLType:valueDef]autorelease];
		else
			*obj = [[[class alloc]init]autorelease];
        
        if(tag&0x08){
            _flags.useChunks = YES; // _setUseChunk:YES ?
            [self decodeChunk];
        }else{
            _flags.useChunks = _flags.inChunk = NO; // must match higher level value, if any
        }
        
        @try{
			[self registerObject:*obj withLocation:location];
			_valueLevel += 1;
			[valueDef decodeObjectState:obj withCoder:self];
			// Original valuetype may want to update
			[valuetype finishDecodingObject:obj withObjectCoder:valueDef];
			[self registerObject:*obj withLocation:location]; // Boxed value can change *obj
			
			// All nested valutypes must be chunked if parent is chunked
			if([self _useChunks] && ![self isAtEOF:sizeof(long)]){
				[self closeChunk]; // ????
				[super _alignCursorTo:sizeof(long)];
				tag = [self _peepTag];
				if(tag < 0 && _valueLevel + tag == 0) // if this is our tag, take it
					_cursor += sizeof(long);
				else
				if(tag < 0 && _valueLevel + tag < 0)
					[NSException raise:CORBAMarshalException
						format:@"incorrect value-end tag, value level mismatch"];
				else if(tag >= 0)
					[NSException raise:CORBAMarshalException format:@"missing value-end tag"];
			}
			
			_valueLevel -= 1;
        }@catch(NSException *localException){
			if([[localException name]isEqualToString:IDLValueEndException]){
				tag = [self _peepTag];
				_valueLevel -= 1;
				if(tag < 0){
					if(_valueLevel + tag >= 0)// value-end-tag belongs to outer value
						[localException raise];
					else// it is our end-tag, take it
						_cursor += sizeof(long);
				}
			}else
				[localException raise];
        }
    }else{ // Invalid value
        [NSException raise:CORBAMarshalException
                format:@"value-tag expected, read '0x%x' while %@in chunk", 
                    tag, _flags.inChunk?@"":@"not "];
    }

}

- (void) encodeEndTag
{
    if(_flags.inChunk){ // No previous end-tag
        [self closeChunk];
        [super _alignTo:sizeof(long)];
        [self _encodeLong:-_valueLevel];
    }else{
        long l = [_data length];
        if(l%sizeof(long) == 0){ // may be previous end-tag
            long tag = [self decodeLongAt:l-sizeof(long)];
            if(tag  == -(_valueLevel + 1)){ // previous end-tag
                [self encodeLong:-_valueLevel at:l-sizeof(long)];
            }else{
                [super _alignTo:sizeof(long)];
                [self _encodeLong:-_valueLevel];
            }
        }else{
            [super _alignTo:sizeof(long)];
            [self _encodeLong:-_valueLevel];
        }
    }
}

- (void) _encodeValue:obj forObjectCoder:(IDLValueType*)valuetype
{
    NSString* typeId = [valuetype typeId];
    long tag = 0x7fffff02;
    long location;
    
    [self closeChunk];
    [super _alignTo:sizeof(long)];
    location = [_data length];
    
    if(_valueLevel == 0){
        if(_encodesWithChunks){
            tag += 0x08;
            _flags.useChunks = YES;
		}else
            _flags.useChunks = NO;
    }else if(_flags.useChunks)
		tag += 0x08;
    
    [super _alignTo:sizeof(long)];
    [self _encodeLong:tag];
    [self _encodeString:typeId];
    
    [self registerLocation:location forObject:obj];

    _valueLevel += 1;
    [valuetype encodeObjectState:obj withCoder:self];
	if(_flags.useChunks) // value-end tag only if use chanks
		[self encodeEndTag];
    _valueLevel -= 1;
    
}
- (void) encodeValue:obj forObjectCoder:(IDLValueType*)valuetype
{
    if(obj){
        long location = [self locationForObject:obj];
        if(location){
            long tag = -1;
            [self encodeLong:tag];
            [self encodeLong:(location - [_data length])];
        }else
            [self _encodeValue:obj forObjectCoder:valuetype];
    }else
        [self encodeLong:0];
}

- (VXMessage*)message
{
	return _message;
}

- (void) setMessage:(VXMessage*)message
{
	_message = message; // mesage owns stream, do not retain it
}

@end

@implementation IDLValueType(ChunkEncoding)
+ (BOOL) encodeValuesUsingChunks
{
	return _encodesWithChunks;
}

+ (void) setEncodeValuesUsingChunks:(BOOL)useChunks
{
	_encodesWithChunks = useChunks? YES : NO;
}

@end

